/*----------------------------------------------------------------------
  File    : ptlx.c
  Contents: prototype-less cluster execution
  Authors : Christian Borgelt
  History : 2007.01.16 file created
            2007.01.19 first version completed
            2007.01.22 alignment of membership columns added
            2007.01.23 read number of objects to consider
            2007.02.27 offset for membership degrees added
            2007.05.04 bug in scaling the data table fixed
            2013.08.20 adapted to definitions ATTID, VALID, TPLID etc.
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_WRITE
#define AS_WRITE
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#include "table.h"
#include "ptlset.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

#ifdef _MSC_VER
#ifndef snprintf
#define snprintf   _snprintf
#endif
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "ptlx"
#define DESCRIPTION "prototype-less cluster set execution"
#define VERSION     "version 2.1 (2014.10.24)         " \
                    "(c) 2007-2014   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_PARSE      (-9)       /* parse errors on input file */
#define E_MATCH     (-10)       /* mismatch of number of rows */
#define E_MSEXP     (-11)       /* invalid membership exponent */
#define E_MSOFF     (-12)       /* invalid membership offset */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- prediction result --- */
  DIMID  cnt;                   /* number of clusters */
  CCHAR *col_cls;               /* name   of cluster    column */
  int    cwd_cls;               /* width  of cluster    column */
  CCHAR *col_msd;               /* name   of membership column(s) */
  int    cwd_msd;               /* width  of membership column(s) */
  int    dig_msd;               /* digits of membership column(s) */
} RESULT;                       /* (prediction result) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_MATCH   -10 */  "mismatch in number of tuples "
                         "(%"DIMID_FMT" vs. %"DIMID_FMT")",
  /* E_MSEXP   -11 */  "invalid membership exponent %g",
  /* E_MSOFF   -12 */  "invalid membership offset %g",
  /*           -13 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR    *prgname;       /* program name for error messages */
static SCANNER  *scan   = NULL; /* scanner */
static TABREAD  *tread  = NULL; /* table reader */
static TABWRITE *twrite = NULL; /* table writer */
static ATTSET   *attset = NULL; /* attribute set */
static TABLE    *ptltab = NULL; /* table (for cluster set) */
static TABLE    *table  = NULL; /* table (for data to assign) */
static PTLSET   *ptlset = NULL; /* prototype-less cluster set */
static TUPLE    *tpl    = NULL; /* tuple for scaling and output */
static double   *scfs   = NULL; /* scaling factors for normalization */
static double   *offs   = NULL; /* offsets for normalization */
static double   *dists  = NULL; /* buffer for distances */
static double   *msdegs = NULL; /* membership degrees */
static RESULT   res     = {     /* prediction result */
  0,                            /* number of clusters */
  NULL, 0,                      /* data for cluster    column */
  NULL, 0, 6 };                 /* data for membership column(s) */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (dists)  free(dists);           \
  if (scfs)   free(scfs);            \
  if (tpl)    tpl_delete(tpl);       \
  if (ptlset) ptl_delete(ptlset);    \
  if (table)  tab_delete(table,  0); \
  if (attset) as_delete(attset);     \
  if (tread)  trd_delete(tread,  1); \
  if (twrite) twr_delete(twrite, 1); \
  if (scan)   scn_delete(scan,   1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static void infout (ATTSET *set, TABWRITE *twrite, int mode)
{                               /* --- write additional information */
  DIMID c;                      /* loop variable for clusters */
  int   n, k, m;                /* character counters */
  char  b[64];                  /* buffer for number output */

  assert(set && twrite);        /* check the function arguments */
  if (res.col_cls) {            /* if to write a cluster index */
    if (mode & AS_ATT) {        /* if to write the header */
      twr_puts(twrite, res.col_cls);  /* write cluster column name */
      if ((mode & AS_ALIGN)     /* if to align the column */
      && ((mode & AS_WEIGHT) || res.col_msd)) {
        n = (int)strlen(res.col_cls);
        k = snprintf(b, sizeof(b), "%"DIMID_FMT, res.cnt);
        res.cwd_cls = k = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
        if (k > n) twr_pad(twrite, (size_t)(k-n));
      }                         /* compute width of prediction column */
      if (res.col_msd) {        /* if to write a membership column */
        twr_fldsep(twrite);     /* write a field separator and */
        twr_puts(twrite, res.col_msd);      /* the column name */
        if ((mode & AS_ALIGN) && (mode & AS_WEIGHT)) {
          n = (int)strlen(res.col_msd);
          k = res.dig_msd +3;   /* compute width of membership column */
          res.cwd_msd = k = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
          if (k > n) twr_pad(twrite, (size_t)(k-n));
        }                       /* pad with blanks if requested */
      } }
    else {                      /* if to write a normal record */
      c = ptl_exec(ptlset, dists, NULL, 0, msdegs);
      n = twr_printf(twrite, "%"DIMID_FMT, c);
      if (res.cwd_cls > n) twr_pad(twrite, (size_t)(res.cwd_cls-n));
      if (res.col_msd) {        /* if to write a membership column */
        twr_fldsep(twrite);     /* write separator and membership */
        n = twr_printf(twrite, "%.*g", res.dig_msd, msdegs[c]);
        if (res.cwd_msd > n) twr_pad(twrite, (size_t)(res.cwd_msd-n));
      }                         /* if to align, pad with blanks */
    } }
  else {                        /* if to write membership degrees */
    if (mode & AS_ATT) {        /* if to write the header */
      if (!res.col_msd) res.col_msd = "";
      m = (int)strlen(res.col_msd);
      if (mode & AS_ALIGN) {    /* if to align the column */
        n = snprintf(b, sizeof(b), "%"DIMID_FMT, res.cnt) +m;
        k = res.dig_msd +3;     /* compute width of activ. columns */
        res.cwd_msd = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
      }
      for (c = 0; c < res.cnt; c++) {
        if (c > 0) twr_fldsep(twrite);
        twr_puts(twrite, res.col_msd);
        n = m +twr_printf(twrite, "%"DIMID_FMT, c+1);
        if (res.cwd_msd > n) twr_pad(twrite, (size_t)(res.cwd_msd-n));
      } }                       /* print a column for each cluster */
    else {                      /* if to write a normal record */
      ptl_exec(ptlset, dists, NULL, 0, msdegs);
      for (c = 0; c < res.cnt; c++) {  /* execute the cluster set */
        if (c > 0) twr_fldsep(twrite); /* and traverse the clusters */
        n = twr_printf(twrite, "%.*g", res.dig_msd, msdegs[c]);
        if (res.cwd_msd > n) twr_pad(twrite, (size_t)(res.cwd_msd-n));
      }                         /* for each cluster print */
    }                           /* the membership degree and */
  }                             /* pad the column for alignment */
}  /* infout() */

/*--------------------------------------------------------------------*/

int main (int argc, char* argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_ptl  = NULL;      /* name of cluster set file */
  CCHAR   *fn_hdr1 = NULL;      /* name of table header file 1 */
  CCHAR   *fn_hdr2 = NULL;      /* name of table header file 2 */
  CCHAR   *fn_tab1 = NULL;      /* name of table file 1 */
  CCHAR   *fn_tab2 = NULL;      /* name of table file 2 */
  CCHAR   *fn_out  = NULL;      /* name of output file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  int     mode1    = AS_ATT|AS_NONULL|AS_NOXATT;
  int     mode2    = AS_ATT|AS_NONULL;
  int     mout     = AS_ATT;    /* table file read and write modes */
  double  msexp    = NAN;       /* membership exponent for adaptation */
  double  msoff    = NAN;       /* membership offset   for adaptation */
  double  tfexp    = NAN;       /* exponent for sum transformation */
  DIMID   clscnt;               /* number of clusters */
  DIMID   rowcnt;               /* number of cluster table rows */
  ATT     *att;                 /* to traverse attributes for marking */
  INST    *inst;                /* to traverse instances for scaling */
  ATTID   m, c;                 /* number of attributes */
  TPLID   n, r;                 /* number of data tuples */
  double  w;                    /* weight of data tuples */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] clsfile [-d|-h hdrfile] ctfile\n"
           "          [-d|-h hdrfile] tabfile outfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-x#      overwrite exponent for membership degrees\n");
    printf("-y#      overwrite offset   for membership degrees\n");
    printf("-X#      overwrite exponent for transformation\n");
    printf("-c#      name of the cluster index field       "
                    "(default: membership degrees)\n");
    printf("-m#      name of the membership field(s)       "
                    "(default: \"%s\")\n", res.col_msd);
    printf("-o#      digits of membership degrees          "
                    "(default: %d)\n",     res.dig_msd);
    printf("-a       align fields in output file           "
                    "(default: single separator)\n");
    printf("-w       do not write field names to output file\n");
    printf("-r#      record  separators                    "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators                    "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                    "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                    "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("clsfile  file to read the cluster description from\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header "
                    "(field names) from hdrfile\n");
    printf("ctfile   cluster table file to read "
                    "(field names in first record)\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("outfile  file to write extended tuples to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'x': msexp  = strtod(s, &s);              break;
          case 'y': msoff  = strtod(s, &s);              break;
          case 'X': tfexp  = strtod(s, &s);              break;
          case 'c': optarg = &res.col_cls;               break;
          case 'm': optarg = &res.col_msd;               break;
          case 'o': res.dig_msd = (int)strtol(s, &s, 0); break;
          case 'w': mout  &= ~AS_ATT;                    break;
          case 'a': mout  |=  AS_ALIGN;                  break;
          case 'r': optarg = &recseps;                   break;
          case 'f': optarg = &fldseps;                   break;
          case 'b': optarg = &blanks;                    break;
          case 'C': optarg = &comment;                   break;
          case 'n': if (k <= 0) mode1 |= AS_WEIGHT;
                    else {      mode2 |= AS_WEIGHT;
                                mout  |= AS_WEIGHT; }    break;
          case 'd': if (k <= 0) mode1 |= AS_DFLT;
                    else {      mode2 |= AS_DFLT; }      break;
          case 'h': optarg    = (k <= 0)
                              ? &fn_hdr1 : &fn_hdr2;     break;
          default : error(E_OPTION, *--s);               break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_ptl  = s;     break;
        case  1: fn_tab1 = s;     break;
        case  2: fn_tab2 = s;     break;
        case  3: fn_out  = s;     break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if (k != 4) error(E_ARGCNT);  /* check number of arguments */
  if (fn_hdr1 && (strcmp(fn_hdr1, "-") == 0))
    fn_hdr1 = "";               /* convert "-" to "" */
  if (fn_hdr2 && (strcmp(fn_hdr2, "-") == 0))
    fn_hdr2 = "";               /* convert "-" to "" */
  i = (!fn_ptl  || !*fn_ptl) ? 1 : 0;
  if  (!fn_tab1 || !*fn_tab1) i++;
  if  (!fn_tab2 || !*fn_tab2) i++;
  if  ( fn_hdr1 && !*fn_hdr1) i++;
  if  ( fn_hdr2 && !*fn_hdr2) i++;
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  mout |= AS_INFO1|AS_RDORD;
  if ((mout & AS_ATT) && (mout & AS_ALIGN))
    mout |= AS_ALNHDR;          /* set align to header flag */
  if (!isnan(msexp) && (msexp <=  0)) error(E_MSEXP, msexp);
  if (!isnan(msoff) && (msoff <= -1)) error(E_MSOFF, msoff);
  if (!isnan(tfexp) && (tfexp <   0)) tfexp = 1.0/(msexp -1.0);
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- parse cluster set --- */
  scan = scn_create();          /* create a scanner */
  if (!scan) error(E_NOMEM);    /* for the cluster set */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_ptl) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  if (as_parse(attset, scan, AT_ALL, 1) != 0)
    error(E_PARSE, scn_name(scan)); /* parse domain descriptions */
  ptlset = ptl_parse(scan);         /* parse the cluster set */
  if (!ptlset || !scn_eof(scan, 1))
    error(E_PARSE, scn_name(scan));
  clscnt = ptl_clscnt(ptlset);  /* get the number of clusters */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", as_attcnt(attset));
  fprintf(stderr, " %"DIMID_FMT" cluster(s)]",   clscnt);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  ptl_msexp(ptlset, msexp, msoff); /* overwrite exponents and */
  ptl_tfexp(ptlset, tfexp);        /* offset with given values */
  res.cnt = clscnt;             /* note the number of clusters */

  /* --- read cluster table header --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread) error(E_NOMEM);   /* set the separator characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  if (fn_hdr1) {                /* if a header file is given */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_hdr1) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, (mode1 & ~AS_DFLT) | AS_ATT);
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_close(tread);           /* read table header, close file */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    mode1 &= ~(AS_ATT|AS_DFLT); /* print a success message and */
  }                             /* remove the attribute flag */

  /* --- read cluster table --- */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_tab1) != 0)
    error(E_FOPEN, trd_name(tread));
  ptltab = tab_create("cluster table", attset, tpl_delete);
  if (!ptltab) error(E_NOMEM);  /* read the data table */
  k = tab_read(ptltab, tread, mode1);
  if (k < 0) error(-k, tab_errmsg(ptltab, NULL, 0));
  trd_delete(tread, 1);         /* delete the table reader */
  tread = NULL;                 /* and clear the variable */
  m = tab_attcnt(ptltab);       /* get the number of attributes */
  n = tab_tplcnt(ptltab);       /* and the number of tuples */
  w = tab_tplwgt(ptltab);       /* and print a success message */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
  if (ptl_rowcnt(ptlset) != n)  /* check the number of tuples */
    error(E_MATCH, ptl_rowcnt(ptlset), n);
  if (mode1 & AS_WEIGHT) {      /* if there are tuple weights */
    for (r = 0; r < n; r++)     /* set them as object weights */
      ptl_setwgt(ptlset, r, tpl_getwgt(tab_tpl(ptltab, r)));
  }
  if (ptl_getnorm(ptlset)) {    /* if to normalize input ranges */
    scfs = (double*)malloc(2 *(size_t)m *sizeof(double));
    if (!scfs) error(E_NOMEM);  /* create buffer vectors for */
    offs = scfs +m;             /* scaling factors and offsets */
    for (c = 0; c < m; c++) {   /* traverse the table columns */
      k = att_type(tab_col(ptltab, c));
      if      (k == AT_INT) tab_colconv(ptltab, c, AT_FLT);
      else if (k != AT_FLT) { scfs[c] = offs[c] = 0; continue; }
      tab_colnorm(ptltab, i, 0, 1, scfs+c, offs+c);
    }                           /* normalize all numeric columns */
  }                             /* to mean 0 and std. deviation 1 */
  dists = (double*)malloc(((size_t)n +(size_t)clscnt) *sizeof(double));
  if (!dists) error(E_NOMEM);   /* create arrays for distances */
  msdegs = dists +(rowcnt = n); /* and membership degrees */
  as_setmark(attset, 1);        /* mark all attributes */

  /* --- read table header --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread) error(E_NOMEM);   /* set the separator characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  if (fn_hdr2) {                /* if a header file is given */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_hdr2) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, (mode2 & ~AS_DFLT) | AS_ATT);
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_close(tread);           /* read table header, close file */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    mode2 &= ~(AS_ATT|AS_DFLT); /* print a success message and */
  }                             /* remove the attribute flag */

  /* --- process table body --- */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_tab2) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  if (mout & AS_ALIGN) {        /* if to align the output columns */
    table = tab_create("table", attset, tpl_delete);
    if (!table) error(E_NOMEM); /* read the data table */
    k = tab_read(table, tread, mode2);
    if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
    trd_delete(tread, 1);       /* delete the table reader */
    tread = NULL;               /* and clear the variable */
    n = tab_tplcnt(table);      /* get the number of tuples */
    w = tab_tplwgt(table);      /* and print a success message */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", tab_attcnt(table));
    fprintf(stderr, " %"TPLID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
    for (c = 0; c < as_attcnt(attset); c++) {
      att = as_att(attset, c);  /* unmark the added attributes */
      if (att_getmark(att) <= 0) att_setmark(att, -1);
    }                           /* (prevent use of add. attributes) */
    tpl = tpl_create(attset, 0);/* create a tuple for the output */
    if (!tpl)    error(E_NOMEM);/* (for scaling and distance comp.) */
    twrite = twr_create();      /* create a table writer and */
    if (!twrite) error(E_NOMEM);/* configure the characters */
    twr_xchars(twrite, recseps, fldseps, blanks, "");
    t = clock();                /* start timer, open output file */
    if (twr_open(twrite, NULL, fn_out) != 0)
      error(E_FOPEN, twr_name(twrite));
    fprintf(stderr, "writing %s ... ", twr_name(twrite));
    if ((mout & AS_ATT)         /* write a table header */
    &&  (as_write(attset, twrite, mout, infout) != 0))
      error(E_FWRITE, twr_name(twrite));
    mout = AS_INST | (mout & ~AS_ATT);
    m += (res.col_cls ? (res.col_msd ? 2 : 1) : res.cnt);
    for (r = 0; r < n; r++) {   /* traverse the tuples */
      tpl_toas(tab_tpl(table, r)); /* copy tuple to attribute set */
      tpl_copy(tpl, tab_tpl(table, r));
      if (scfs) {               /* if to scale the data */
        for (c = m; --c >= 0;){ /* traverse the numeric columns */
          if (scfs[c] <= 0) continue;
          inst = tpl_colval(tpl, c);
          inst->f = (DTFLT)((double)inst->f *scfs[c] +offs[c]);
        }                       /* scale the numeric columns */
      }                         /* and write the current tuple */
      dv_fromtab(dists, rowcnt, NULL, 0, ptltab, tpl);
      if (as_write(attset, twrite, mout, infout) != 0)
        error(E_FWRITE, twr_name(twrite));
    } }                         /* write the current tuple */
  else {                        /* if to process tuples directly */
    k = as_read(attset, tread, mode2);  /* read the table header */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    for (c = 0; c < as_attcnt(attset); c++) {
      att = as_att(attset, c);  /* unmark the added attributes */
      if (att_getmark(att) <= 0) att_setmark(att, -1);
    }                           /* (prevent use of add. attributes) */
    tpl = tpl_create(attset, 0);/* create a tuple for the output */
    if (!tpl)    error(E_NOMEM);/* (for scaling and distance comp.) */
    twrite = twr_create();      /* create a table writer and */
    if (!twrite) error(E_NOMEM);/* configure the characters */
    twr_xchars(twrite, recseps, fldseps, blanks, "");
    t = clock();                /* start timer, open output file */
    if (twr_open(twrite, NULL, fn_out) != 0)
      error(E_FOPEN, twr_name(twrite));
    if ((mout & AS_ATT)         /* write a table header */
    &&  (as_write(attset, twrite, mout, infout) != 0))
      error(E_FWRITE, twr_name(twrite));
    mout = AS_INST | (mout & ~AS_ATT);
    i = mode2; mode2 = (mode2 & ~(AS_DFLT|AS_ATT)) | AS_INST;
    if (!(i & AS_DFLT))         /* if not done yet, read first tuple */
      k = as_read(attset, tread, mode2);
    for (w = 0, n = 0; k == 0; n++) {
      tpl_fromas(tpl);          /* traverse the tuples */
      if (scfs) {               /* if to scale the data */
        for (c = m; --c >= 0;){ /* traverse the numeric columns */
          if (scfs[c] <= 0) continue;
          inst = tpl_colval(tpl, c);
          inst->f = (DTFLT)((double)inst->f *scfs[c] +offs[c]);
        }                       /* scale the numeric columns */
      }                         /* and write the current tuple */
      dv_fromtab(dists, rowcnt, NULL, 0, ptltab, tpl);
      if (as_write(attset, twrite, mout, infout) != 0)
        error(E_FWRITE, twr_name(twrite));
      w += as_getwgt(attset);   /* sum the tuple weight */
      k = as_read(attset, tread, mode2);
    }                           /* try to read the next tuple */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_delete(tread, 1);       /* delete the table reader */
    tread = NULL;               /* and clear the variable */
    m = as_attcnt(attset);      /* get the number of attributes */
  }
  if (twr_close(twrite) != 0) error(E_FWRITE, twr_name(twrite));
  twr_delete(twrite, 1);        /* close the output file and */
  twrite = NULL;                /* delete the table writer */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
